/*
 * Copyright (c) 2015 by inovat, innovative systeme - verkehr - tunnel - technik, Dipl.-Ing. H. C. Kniss
 *
 * This file is part of de.bsvrz.kex.tls.osi7.Device
 *
 * de.bsvrz.kex.tls.osi7.Device is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  de.bsvrz.kex.tls.osi7.Device is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with de.bsvrz.kex.tls.osi7.Device.  If not, see <http://www.gnu.org/licenses/>.
 *
 *  Diese Datei ist Teil von de.bsvrz.kex.tls.osi7.Device.
 *
 *  de.bsvrz.kex.tls.osi7.Device ist Freie Software: Sie können es unter den Bedingungen
 *  der GNU General Public License, wie von der Free Software Foundation,
 *  Version 3 der Lizenz oder (nach Ihrer Wahl) jeder späteren
 *  veröffentlichten Version, weiterverbreiten und/oder modifizieren.
 *
 *  de.bsvrz.kex.tls.osi7.Device wird in der Hoffnung, dass es nützlich sein wird, aber
 *  OHNE JEDE GEWÄHRLEISTUNG, bereitgestellt; sogar ohne die implizite
 *  Gewährleistung der MARKTFÄHIGKEIT oder EIGNUNG FÜR EINEN BESTIMMTEN ZWECK.
 *  Siehe die GNU General Public License für weitere Details.
 *
 *  Sie sollten eine Kopie der GNU General Public License zusammen mit diesem
 *  Programm erhalten haben. Wenn nicht, siehe <http://www.gnu.org/licenses/>.
 *
 * Contact Information:
 * inovat, Dipl.-Ing. H. C. Kniss
 * An der Krautwiese 37
 * D-53783 Eitorf
 * +49 (0)2243 8464 193
 * info@inovat.de
 * www.inovat.de
 */

package de.bsvrz.kex.tls.osi7;

import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.dav.daf.main.config.*;
import de.bsvrz.kex.tls.osi2osi3.osi3.NetworkLayerListener;
import de.bsvrz.kex.tls.osi7.conversion.DeviceConversionInterface;
import de.bsvrz.sys.funclib.debug.Debug;

import java.lang.reflect.InvocationTargetException;
import java.util.Map;
import java.util.regex.Pattern;

/**
 * TBD: Kurzbeschreibung TBD: Langebeschreibung
 *
 * @author inovat, innovative systeme - verkehr - tunnel - technik
 * @author Dipl.-Ing. Hans Christian Kniß (HCK)
 * @version $Revision$ / $Date$ / ($Author$)
 */
public class Device {
	// ------------------------------------------------------------------------
	// VARIABLENDEKLARATIONEN
	// ------------------------------------------------------------------------

	// Sonstige ---------------------------------------------------------------
	/** DebugLogger für Debug-Ausgaben */
	private static final Debug debug = Debug.getLogger();

	/** Instanz des UmsetzungsModuls für die Konvertierung der Telegramme in Attributgruppen (und zurück) */
	DeviceConversionInterface _conversionModul;
	// Private ----------------------------------------------------------------
	/** Klassenobjekt des UmsetzungsModuls für die Konvertierung der Telegramme in Attributgruppen (und zurück) */
	private Class _conversionClass;

	/** DaV Interface */
	private final ClientDavInterface _daf;

	/** Objektreferenz auf das Objekt vom "typ.gerät" aus dem Teilmodell TLS. */
	private final ConfigurationObject _device;

	/** Name des UmsetzungsModul für die Konvertierung der Telegramme in Attributgruppen (und zurück) */
	private String _conversionModulName;

	/** Bezeichnung des Gerätes. */
	private final String _deviceDescription;

	/** Verbindungsstatus. */
	private int _connectionState = NetworkLayerListener.DEVICE_DISCONNECTED;

	/** OSI7 KnotenNummer. */
	private final int _osi7Adress;

	// ------------------------------------------------------------------------
	// KONSTRUKTOREN  (und vom Konstruktor verwende Methoden)
	// ------------------------------------------------------------------------

	/**
	 * Erzeugt eine Instanz eines Device (Stellvertreterobjekt für typ.gerät).
	 *
	 * @param daf DaV.
	 * @param device Device
	 */
	public Device(ClientDavInterface daf, ConfigurationObject device) {
		_device = device;
		_daf = daf;

		DataModel config = _daf.getDataModel();
		AttributeGroup devicePropertiesAtg = config.getAttributeGroup("atg.gerät");
		Data deviceProperties = _device.getConfigurationData(devicePropertiesAtg);

		_deviceDescription = deviceProperties.getTextValue("Bezeichnung").getText();
		_osi7Adress = deviceProperties.getScaledValue("KnotenNummer").intValue();
		_conversionModulName = deviceProperties.getTextValue("UmsetzungsModul").getText().trim();

		installConversionModule();
	}

	private void installConversionModule() {
		try {
			if (_conversionModulName.equals("")) {
				_conversionModulName = "de.bsvrz.kex.tls.osi7.conversion.DeviceDefault";
			}

			// Anpassung zur Kompatibilität nach Umstellung auf neue Packagestruktur de.bsvrz.*
			// Liegt Umsetzungsklasse noch im alten Package kex.tls.osi7.conversion.*, dann
			// wird automatisch im Package de.bsvrz.kex.tls.osi7.conversion.* gesucht.
			debug.fine("KonvertierungsModulName Vorgabe: [" + _conversionModulName + "]");
			if (Pattern.matches("^kex\\.tls\\.osi7\\.conversion\\..*", _conversionModulName)) {
				_conversionModulName = ("de.bsvrz." + _conversionModulName).trim();

			}
			debug.fine("KonvertierungsModulName nach Anpassung auf neue Packagestruktur de.bsvrz.*: [" + _conversionModulName + "]");

			// Umwandlungsmodul laden und instanzieren
			debug.finest("Versuche Umsetzungsklasse zu laden: " + _conversionModulName);

			_conversionClass = Class.forName(_conversionModulName);
			_conversionModul = (DeviceConversionInterface) _conversionClass.getConstructor(new Class[]{ClientDavInterface.class, Device.class}).newInstance(new Object[]{_daf, this});
		}
		catch (ClassNotFoundException e) {
			debug.error("Konnte UmsetzungsModul [" + _conversionModulName + "] nicht laden!", e);
		}
		catch (NoSuchMethodException e) {
			debug.error("Konstruktoraufruf in Klasse[" + _conversionModulName + "] fehlgeschlagen!", e);
		}
		catch (IllegalAccessException e) {
			debug.error("Zugriff auf UmsetzungsModul [" + _conversionModulName + "] fehlgeschlagen!", e);
		}
		catch (InstantiationException e) {
			debug.error("UmsetzungsModul [" + _conversionModulName + "] konnte  nicht instanziert werden!", e);
		}
		catch (InvocationTargetException e) {
			debug.error("Konstruktoraufruf am UmsetzungsModul [" + _conversionModulName + "] fehlgeschlagen!", e);
		}
	}

	/**
	 * Erzeugt eine Instanz eines Device für die TLS ähnliche Kommunikaton (Stellvertreterobjekt für ein Steuermodul).
	 *
	 * @param daf DaV
	 * @param osi7Adress osi7Adress
	 * @param conversionModules conversionModules
	 */
	public Device(ClientDavInterface daf, String osi7Adress, Map conversionModules) {
		_daf = daf;
		//noinspection AssignmentToNull
		_device = null;

		DataModel config = _daf.getDataModel();

		_deviceDescription = "Virtuelles SM für TLS ähnliche Kommunikation";
		_osi7Adress = Integer.parseInt(osi7Adress);

		try {
			_conversionModulName = ((String) conversionModules.get(config.getObject("typ.steuerModul"))).trim();
		}
		catch (NullPointerException e) {
			debug.error("UmsetzungsModuel für SM (bei TLS ähnlicher Kommunikation) muss \n" + "in [atg.gerätTlsÄhnlicheKommunikationsPartnerUmsetzungsModule] \n" + "einmal für den Typ [typ.steuerModul] für das Gerät (VIZ, VRZ, UZ) angegeben werden!", e);
		}

		installConversionModule();
	}

	// ------------------------------------------------------------------------
	// GETTER / SETTER METHODEN
	// ------------------------------------------------------------------------

	/**
	 * Bestimmt das diesem Gerät zugeordnete Konfigurationsobjekt vom "typ.gerät" aus dem Teilmodell TLS.
	 *
	 * @return Das diesem Gerät zugeordnete Konfigurationsobjekt.
	 */
	public ConfigurationObject getDevice() {
		return _device;
	}

	/**
	 * Liefert die Bezeichungstext des Gerätes.
	 *
	 * @return Bezeichnungstext des Gerätes
	 */
	public String getDeviceDescription() {
		return _deviceDescription;
	}

	/**
	 * Liefert einen Hashkey auf Basis der KnotenNummer als String in Dezimaldarstellung (0..16777215).
	 *
	 * @return Hashkey auf Basis der KnotenNummer
	 */
	String getKey() {
		return (String.valueOf(_osi7Adress));
	}

	/**
	 * Liefert die KnotenNummer des Gerätes.
	 *
	 * @return KnotenNummer des Gerätes
	 */
	public int getOsi7Adress() {
		return _osi7Adress;
	}

	/**
	 * Liefert die KnotenNummer des Gerätes in Dezimaldarstellung (0..16777215).
	 *
	 * @return KnotenNummer des Gerätes in Dezimaldarstellung (0..16777215)
	 */
	public String getOsi7AdressDecimal() {
		return String.valueOf(_osi7Adress);
	}

	/**
	 * Liefert die KnotenNummer des Osi7Telegramms als Hex-Darstellung (0..ffffff).
	 *
	 * @return KnotenNummer des Osi7Telegramms als Hex-Darstellung (0..ffffff).
	 */
	public String getOsi7AdressHex() {
		return Integer.toHexString(_osi7Adress);
	}

	/**
	 * Liefert die KnotenNummer des Osi7Telegramms in der Darstellung "LocationCode-Distanz" (0-0 .. 65535-255).
	 *
	 * @return KnotenNummer des Osi7Telegramms in der Darstellung "LocationCode-Distanz" (0-0 .. 65535-255).
	 */
	public String getOsi7AdressLocCode() {
		return ((_osi7Adress >> 8) + "-" + (_osi7Adress & 0xff));
	}

	/**
	 * Liefert den internen Verbindungsstatus des Geräts.
	 *
	 * @return Status der Verbindung ({@link NetworkLayerListener#DEVICE_CONNECTED}, {@link
	 *         NetworkLayerListener#DEVICE_DISCONNECTED}) .
	 */
	public int getState() {
		return _connectionState;
	}

	/**
	 * Setzt den internen Verbindungsstatus des Geräts.
	 *
	 * @param state Status der Verbindung ({@link @kex.tls.osi3.NetworkLayerListener.DEVICE_CONNECTED}) .
	 */
	void setState(int state) {
		if (state == NetworkLayerListener.DEVICE_CONNECTED) {
			_connectionState = state;
		}
		else if (state == NetworkLayerListener.DEVICE_DISCONNECTED) {
			_connectionState = state;
		}
		else {
			debug.warning("Unbekannten VerbindungsStatus für Knoten " + this.getOsi7AdressDecimal() + " erhalten: [" + state + "].");
		}
	}

	// ------------------------------------------------------------------------
	// KANONISCHE METHODEN
	// ------------------------------------------------------------------------

	/**
	 * Liefert die String-Darstellung eines Gerätes:
	 * <pre>
	 * Gerät (Bezeichnung)             :
	 * Gerätetyp                       :
	 * Knotennummer (Dez:Hex:Loc-Code) :
	 * Umsetzungsmodul                 :
	 * Schlüsselwert (HashKey)         :
	 * Verbindungsstatus               :
	 * </pre>
	 *
	 * @return String-Darstellung
	 */
	public String toString() {
		StringBuffer sb = new StringBuffer();
		final String NEWLINE = System.getProperty("line.separator");
		sb.append("................................................................................").append(NEWLINE);
		sb.append("Gerät (Bezeichnung)             : ").append(getDeviceDescription()).append(NEWLINE);
		sb.append("Gerätetyp                       : ").append(getObjectType()).append(NEWLINE);
		sb.append("Knotennummer (Dez:Hex:Loc-Code) : ").append(getOsi7AdressDecimal()).append(":").append(getOsi7AdressHex()).append(":").append(getOsi7AdressLocCode()).append(NEWLINE);
		sb.append("Umsetzungsmodul                 : ").append(_conversionModul).append(NEWLINE);
		sb.append("Schlüsselwert (HashKey)         : ").append(getKey()).append(NEWLINE);
		sb.append("Verbindungsstatus               : ").append(getState() == NetworkLayerListener.DEVICE_CONNECTED ? "Verbunden" : "Nicht verbunden").append(NEWLINE);
		return (sb.toString());
	}

	// ------------------------------------------------------------------------
	// SONSTIGE METHODEN
	// ------------------------------------------------------------------------

	/**
	 * Liefert Namen/Pid/Id des Objekttyps diese Gerätes.
	 *
	 * @return Objekttyps diese Gerätes
	 */
	String getObjectType() {
		String s = "Virtuelles SM für TLS ähnliche Kommunikation";
		if (_device != null) {
			s = _device.getType().getNameOrPidOrId();
		}
		return s;
	}

	/**
	 * Verarbeitet ein Osi7Telegramm.
	 *
	 * @param osi7Telegram Das zu verarbeitende {@link Osi7Telegram}.
	 */
	void processOsi7Telegram(Osi7Telegram osi7Telegram) {
		// Zuerst spezielle Verarbeitung entsprechend dem geladenen Umsetzungsmodul durchführen...
		_conversionModul.processOsi7Telegram(osi7Telegram);
		// ...danach Telegramm an die nächste Verarbeitungsstufe weitergeben.
		// Bei normalen Geräten derzeit keine Weiterverarbeitung, bei Sm weitergabe der
		// EinzelTelegramme an die zuständigen Eak.
	}
}
