/*
 * Copyright (c) 2015 by inovat, innovative systeme - verkehr - tunnel - technik, Dipl.-Ing. H. C. Kniss
 *
 * This file is part of de.bsvrz.kex.tls.osi7.conversion.nw.vrz.Fg001HeuschBoesefeldt
 *
 * de.bsvrz.kex.tls.osi7.conversion.nw.vrz.Fg001HeuschBoesefeldt is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  de.bsvrz.kex.tls.osi7.conversion.nw.vrz.Fg001HeuschBoesefeldt is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with de.bsvrz.kex.tls.osi7.conversion.nw.vrz.Fg001HeuschBoesefeldt.  If not, see <http://www.gnu.org/licenses/>.
 *
 *  Diese Datei ist Teil von de.bsvrz.kex.tls.osi7.conversion.nw.vrz.Fg001HeuschBoesefeldt.
 *
 *  de.bsvrz.kex.tls.osi7.conversion.nw.vrz.Fg001HeuschBoesefeldt ist Freie Software: Sie können es unter den Bedingungen
 *  der GNU General Public License, wie von der Free Software Foundation,
 *  Version 3 der Lizenz oder (nach Ihrer Wahl) jeder späteren
 *  veröffentlichten Version, weiterverbreiten und/oder modifizieren.
 *
 *  de.bsvrz.kex.tls.osi7.conversion.nw.vrz.Fg001HeuschBoesefeldt wird in der Hoffnung, dass es nützlich sein wird, aber
 *  OHNE JEDE GEWÄHRLEISTUNG, bereitgestellt; sogar ohne die implizite
 *  Gewährleistung der MARKTFÄHIGKEIT oder EIGNUNG FÜR EINEN BESTIMMTEN ZWECK.
 *  Siehe die GNU General Public License für weitere Details.
 *
 *  Sie sollten eine Kopie der GNU General Public License zusammen mit diesem
 *  Programm erhalten haben. Wenn nicht, siehe <http://www.gnu.org/licenses/>.
 *
 * Contact Information:
 * inovat, Dipl.-Ing. H. C. Kniss
 * An der Krautwiese 37
 * D-53783 Eitorf
 * +49 (0)2243 8464 193
 * info@inovat.de
 * www.inovat.de
 */

package de.bsvrz.kex.tls.osi7.conversion.nw.vrz;

import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.dav.daf.main.DataDescription;
import de.bsvrz.dav.daf.main.ResultData;
import de.bsvrz.dav.daf.main.config.IntegerValueState;
import de.bsvrz.kex.tls.osi7.De;
import de.bsvrz.kex.tls.osi7.DeBlock;
import de.bsvrz.kex.tls.osi7.Eak;
import de.bsvrz.kex.tls.osi7.TlsApplicationLayer;
import de.bsvrz.sys.funclib.debug.Debug;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * Umsetungsmodul für die FG 1, wird für die WanCom-Kopplung der ZDF basierten VRZ Leverkusen und VRZ Recklinghausen mit
 * der DaV basierten VRZ NRW verwendet.
 *
 * @author inovat, innovative systeme - verkehr - tunnel - technik
 * @author Dipl.-Ing. Hans Christian Kniß (HCK)
 * @version $Revision$ / $Date$ / ($Author$)
 */

public class Fg001HeuschBoesefeldt extends FgAllHeuschBoesefeldt {
    // ------------------------------------------------------------------------
    // VARIABLENDEKLARATIONEN
    // ------------------------------------------------------------------------

    // Sonstige ---------------------------------------------------------------
    /** DebugLogger für Debug-Ausgaben */
    private static final Debug debug = Debug.getLogger();
    // Private ----------------------------------------------------------------
    /**
     * Tabelle, in der das Verhalten beim Empfang von Attributgruppen eingetragen wird. Jede Tabellenzeile hat folgenden
     * Aufbau:
     * <pre>
     * "Attributgruppe (Pid)", "Aspekt (Pid)", "Fg", "Typ", "Id", "ObjektTyp"
     * z.B.
     * "atg.tlsGloKanalSteuerung", "asp.parameterSoll", "1", "29", "2",
     * "typ.deLve"
     * </pre>
     * Die Auswertung erfolgt dabei wie folgt: <ul> <li> Für jede Attributguppen/Aspekt Kombination erfolgt eine
     * Empfangsanmeldung als Empfänger.</li> <li> In der Methode {@link #update} wird für jede zur
     * Attributgruppen/Aspket-Kombination passende Kombination eine Methode aufgerufen, deren Namen aus dem
     * Tabelleneintrag wie folgt gebildet wird (hier für den Beispieleintrag):<br> <code>Fg001Typ029Id002AtgTlsGloKanalSteuerungAspParameterSoll(ResultData)</code><br>
     * Dabei enthält <code>ResultData</code> die Daten der zu verarbeitenden ATG/ASP-Kombination.</li> <li> Für jede
     * Attributguppen/Aspekt Kombination erfolgt eine Empfangsanmeldung als Empfänger. Sind mehrere passende Eintrage
     * vorhanden, werden entsprechend mehrere Methodenaufrufe durchgeführt. Dadurch können mehrere ATG/ASP-Kombinationen
     * in einen De-Block abgebildet werden (z.B. protokollnahe TLS-Attributgruppe und Attributgruppe eines
     * Fachmodells):</li>
     * <li> Ein Eintrag von "255" entspricht dabei einer "Wildcard" (passt immer). </ul>
     */
    private final Object[][] davToTls = new Object[][] {};

    /**
     * Tabelle, in der das Verhalten beim Empfang von TLS De-Blöcken eingetragen wird. Jede Tabellenzeile hat folgenden
     * Aufbau:
     * <pre>
     * "Attributgruppe (Pid)", "Aspekt (Pid)", "Fg", "Typ", "Id", "ObjektTyp"
     * z.B.
     * "atg.tlsGloKanalSteuerung", "asp.tlsAntwort", "1", "29", "2", "typ.deLve"
     * </pre>
     * Die Auswertung erfolgt dabei wie folgt: <ul> <li> Für jede Attributguppen/Aspekt Kombination erfolgt eine
     * Sendeanmeldung als Quelle.</li> <li> In der Methode {@link #processDeBlock} wird für jede zum De-Block passende
     * Fg/Id/Typ-Kombination eine Methode aufgerufen, deren Namen aus dem Tabelleneintrag wie folgt gebildet wird (hier
     * für den Beispieleintrag):<br> <code>Fg001Typ029Id002AtgTlsGloKanalSteuerungAspTlsAntwort(DeBlock, Data,
     * DataDescription)</code><br> Dabei enthält der <code>deBlock</code> die Daten des zu verarbeitenden De-Blocks und
     * <code>data</code> ein Datenobjekt der passenden Attributgruppen/Aspekt-Kombination, welchem direkt die Inhalte
     * des
     * De-Blocks zugewiesen werden können.<code>DataDescription</code> enthält die passende {@link DataDescription} für
     * die zu sendenden Daten.</li> <li> Für jede Attributguppen/Aspekt Kombination erfolgt eine Sendeanmeldung als
     * Quelle. Sind mehrere passende Eintrage vorhanden, werden entsprechend mehrere Methodenaufrufe durchgeführt.
     * Dadurch kann ein De-Block mehrfach ausgewert und an verschiedene Attributgruppen weitergerleitet werden (z.B.
     * protokollnahe TLS-Attributgruppe und Attributgruppe eine Fachmodells):</li> <li> Ein Eintrag von "255" entspricht
     * dabei einer "Wildcard" (passt immer). </ul>
     */
    private final Object[][] tlsToDav = new Object[][] {
        // Attributgruppe, Aspekt, Fg, Typ, Id, ObjektTyp
        {"atg.tlsGloDeFehler", "asp.tlsAntwort", "1", "1", "1", "typ.deLve"}, {"atg.tlsGloKanalSteuerung", "asp.parameterIst", "255", "29", "2", "typ.deLve"}, {"atg.tlsGloGeographischeKenndaten", "asp.tlsZustand", "255", "36", "3", "typ.deLve"}, {"atg.tlsLveBetriebsParameter", "asp.parameterIst", "1", "32", "3", "typ.deLve"}, {"atg.verkehrsDatenKurzZeitMq", "asp.analyseHB", "1", "128", "4", "typ.messQuerschnitt"}, {"atg.störfallZustand", "asp.störfallVerfahrenNRWHB", "1", "144", "4", "typ.fahrStreifen"}, {"atg.störfallZustand", "asp.störfallVerfahrenNRWHB", "1", "152", "4", "typ.messQuerschnitt"},
        // Störfallverfahren UNI Duisburg fehlen noch
        {"atg.verkehrsDatenKurzZeitIntervall", "asp.messWertErsetzung", "1", "160", "4", "typ.fahrStreifen"}, {"atg.tlsLveErgebnisMeldungVersion0Bis4", "asp.tlsAntwort", "1", "160", "4", "typ.deLve"}, {"atg.störfallZustand", "asp.störfallVerfahrenRDSHB", "1", "182", "4", "typ.straßenSegment"},
        // RDS Meldung und Rückmeldung fehlen noch
        {"atg.störfallZustand", "asp.störfallVerfahrenRDSHB", "1", "250", "4", "typ.messQuerschnitt"}, {"atg.verkehrsDatenLangZeitIntervall", "asp.externeErfassung", "1", "204", "4", "typ.fahrStreifen"}};

    /**
     * Intervallzeit Kurzzeitdaten, wird bei HB nicht im Intervallzeitstempel kodiert, deshalb Parameter aus
     * Betriebsparamtertelegramm auslesen
     */
    private int _intervalTime = 60;// initial auf 60 Sekunden setzten

    /**
     * Art der Mittelwertbildung, die bei diesem De eingestellt ist. Wird sowohl beim Setzen der Betriebs- parameter als
     * auch beim Empfang der Rückmeldung (sofern diese erfolgt) angepasst. -1 = unbekannt 0 = gleitende
     * Mittelwertbildung 1 = arithmetische Mittelwertbildung
     */
    private int _kindOfAverage = -1;// initial auf undefiniert setzten

    // ------------------------------------------------------------------------
    // KONSTRUKTOREN  (und vom Konstruktor verwende Methoden)
    // ------------------------------------------------------------------------

    /**
     * Erzeugt ein Objekt vom Typ Fg001Default
     *
     * @param daf DaV
     * @param eak Eak
     * @param de  De
     */
    public Fg001HeuschBoesefeldt(ClientDavInterface daf, Eak eak, De de) {
        super(daf, eak, de);
        subscribeAllData(tlsToDav, davToTls);
    }

    // ------------------------------------------------------------------------
    // GETTER / SETTER METHODEN
    // ------------------------------------------------------------------------

    public int getIntervallTime() {
        return _intervalTime;
    }

    public int getKindOfAverage() {
        return _kindOfAverage;
    }

    // ------------------------------------------------------------------------
    // SONSTIGE METHODEN
    // ------------------------------------------------------------------------

    /**
     * FG 1 : Typ: 1 Id: 1 Name: DE-Fehler (Antwort). Ergänzt globale Behandlung De-Fehler in {@link
     * #Fg255Typ001Id001AtgTlsGloDeFehlerAspTlsAntwort}.
     *
     * @param deBlock         Der zu bearbeitende De-Block.
     * @param data            Der passende Datensatz für die Ziel Attributgruppe/Aspekt-Kombination.
     * @param dataDescription Die Datenbeschreibung für die Attributgruppe/Aspekt-Kombination.
     *
     * @return Ergebnisdatensatz mit den vom DeBlock in die entsprechende Attributgruppe konvertierten Daten.
     */
    public ResultData Fg001Typ001Id001AtgTlsGloDeFehlerAspTlsAntwort(DeBlock deBlock, Data data, DataDescription dataDescription) {
        ResultData resultData = super.Fg255Typ001Id001AtgTlsGloDeFehlerAspTlsAntwort(deBlock, data, dataDescription);

        IntegerValueState DEFehlerStatus = data.getUnscaledValue("DEFehlerStatus").getState();
        if (DEFehlerStatus != null) {
            debug.fine("DEFehlerStatus: " + DEFehlerStatus.getName());
            if (DEFehlerStatus.getName().equals("ok")) {
                // Bei der Kommunikation mit der VRZ werden keine Daten bei De=ok versendet.
                //				sendInitialDeBlock(deBlock.getKey(), 29, 2); // LVE-Kanalsteuerung senden
                //				sendInitialDeBlock(deBlock.getKey(), 32, 3); // LVE-Betriebsparameter senden
            }
            else {
                List deList = new ArrayList();
                deList.add(_de);
                TlsApplicationLayer.writeNoDataToUnreachableDe(deList, _eakConversionModul.getTimeStamp(), -1, "atg.tlsGloDeFehler");
            }
        }

        return resultData;
    }

    /**
     * FG 1 : Typ: 32 Id: 3 Name: LVE-Betriebsparameter (Abruf / Antwort) TLS-Datenbeschreibung:
     * <pre>
     * Position         Bezeichnung                                Erläuterung
     * Byte 1           Länge DE-Block                             Länge des
     * folgenden DE-Blocks [ 10 ]
     * Byte 2           Daten-Endgeräte-Kanal                      [1..254, 255]
     * Byte 3           Typ der DE-Daten                           [ 32 ]
     * Byte 4           Datenversion/Kurzzeitdaten                 [ 0.. 4, 255 =
     * aus ]
     * Byte 5           Erfassungsintervalldauer/Kurzzeitdaten     [
     * 1,2,4,8,12...240, Einheit: 15 s ]
     * Byte 6           Datenversion/Langzeitdaten                 [ 10.. 23, 255
     * =
     * aus ]
     * Byte 7           Erfassungsintervalldauer/Langzeitdaten     siehe unten
     * Byte 8           Glättungsparameter                         Alpha 1 Alpha 1
     * = (Byteinhalt + 1) / 256
     * Byte 9           Glättungsparameter                         Alpha 2 Alpha 2
     * = (Byteinhalt + 1) / 256
     * Byte 10          Längengrenzwert
     * Lkw/Pkw-Unterscheidung 4m+[0..254cm]
     * Byte 11          Art / Startwert für gl. Mittelwertbildung  siehe unten
     * </pre>
     * Inhalt Byte: Datenversion / Kurzzeitdaten<br> Mit Hilfe dieser Bytes wird die Datenversion ausgewählt, mit der
     * die zyklisch erfassten Verkehrsdaten übertragen werden sollen. Bei Geräten, die die Version 4 nicht unterstützen,
     * wird eine 4 in diesem Byte mit einer negativen Quittung (Fehlerursache 3) beantwortet. Wenn auf die Versionen 4
     * umgeschaltet wird, ohne dass zuvor mit dem DEBlock Typ 37 eine entsprechende Parametrierung der Grenzen der
     * Geschwindigkeitsklassen vorgenommen wurde, dann werden die dort beschriebenen Defaultwerte für die Grenzen der
     * Geschwindigkeitsklassen verwendet.<br>
     * <p>
     * Inhalt Byte: Erfassungsintervalldauer / Kurzzeitdaten:<br> Das Byte enthält den Wert der Erfassungsintervalldauer
     * in 15-Sekunden-Einheiten. Es sind nur Werte zugelassen, die ein ganzzahliges Vielfaches haben, welches 60 Minuten
     * ergibt (15s, 30s, 1, 2, 3, 4, 5, 6, 10, 12, 15, 20, 30, 60 Minuten).
     * <p>
     * Inhalt Byte: Datenversion / Langzeitdaten<br> Mit Hilfe dieser Bytes wird die Datenversion ausgewählt, mit der
     * die zyklisch erfassten Verkehrsdaten übertragen werden sollen. Geräte, von denen eine nicht implementierte
     * Version gefordert wird, antworten mit einer negativen Quittung (Fehlerursache 3). Wenn auf die Versionen 17, 18,
     * 22 oder 23 umgeschaltet wird, ohne dass zuvor mit dem DE-Block Typ 38 eine entsprechende Parametrierung der
     * Grenzen der Geschwindigkeitsklassen vorgenommen wurde, dann werden die dort beschriebenen Defaultwerte für die
     * Grenzen der Geschwindigkeitsklassen verwendet.<br>
     * <p>
     * Inhalt Byte: Erfassungsintervalldauer / Langzeitdaten<br> Das Byte enthält den Wert der Erfassungsintervalldauer
     * in Stunden-Einheiten:<br>
     * <pre>
     * Bit7 = 1:    Einheit = Stunden
     * Bit0..6:     Zeitwert in Stunden [1..24]
     * </pre>
     * Es sind nur Stundenwerte zugelassen, die ganzzahlige Teiler oder ganzzahlige Mehrfache von 24 Stunden sind (1, 2,
     * 3, 4, 6, 8, 12, 24 Stunden).<br>
     * <p>
     * Inhalt Byte: Art / Startwert für gleitende Mittelwertbildung:<br> Die gleitende Mittelwertbildung wird nur bei
     * Kurzzeitdaten angewandt. Bei Langzeitdaten wird der arithmetische Mittelwert verwendet.<br>
     * <pre>
     * Bit 0..6:    Startwert der Geschwindigkeit für gleitende Mittelwertbildung
     * [0..127km/h]
     * Bit 7:       Art der Geschwindigkeits-Mittelwertbildung
     *              0=gleitende Mittelwertbildung
     *              1=arithmetische Mittelwertbildung
     * Wichtig:     Bei Auswahl der Version 3 oder 4 müssen immer die
     * Glättungsparameter und Startwerte angegeben werden,
     *              da in diesen Telegrammversionen zusätzlich der gleitende
     * Mittelwert für Steuerungszwecke unabhängig
     *              von der Einstellung in Bit 7 übertragen wird.
     * </pre>
     * Inhalt Byte: Längengrenzwert<br> Bei E/A-Konzentratoren ohne Längenauswertung wird bei der Zuweisung das Byte
     * ignoriert. In der Antwort dieses E/A-Konzentrators wird es auf 255 gesetzt.<br>
     *
     * @param deBlock         Der zu bearbeitende De-Block.
     * @param data            Der passende Datensatz für die Ziel Attributgruppe/Aspekt-Kombination.
     * @param dataDescription Die Datenbeschreibung für die Attributgruppe/Aspekt-Kombination.
     *
     * @return Ergebnisdatensatz mit den vom DeBlock in die entsprechende Attributgruppe konvertierten Daten.
     */
    public ResultData Fg001Typ032Id003AtgTlsLveBetriebsParameterAspParameterIst(DeBlock deBlock, Data data, DataDescription dataDescription) {
        ResultData resultData = null;

        // VersionKurz-Byte aus Telegramm ------------------------------------------------------------------------------
        int _VersionKurz_ = deBlock.getDeData().readUnsignedByteAsInt();

        switch (_VersionKurz_) {
            case 0:
            case 1:
            case 2:
            case 3:
            case 4:
            case 255:
                data.getUnscaledValue("VersionKurzZeitDaten").set(_VersionKurz_);
                break;
            default:
                debug.warning("VersionKurzZeitDaten ungültig : " + _VersionKurz_ + Debug.NEWLINE + "De-Block (Osi7Adresse-DeNummer-FG-Typ-Id): " + deBlock.getLongKey() + " verworfen:" + Debug.NEWLINE + deBlock.toString());
                return resultData;
        }

        // Intervalldaten Kurzzeit-Byte aus Telegramm ------------------------------------------------------------------
        int _IntervallKurz_ = deBlock.getDeData().readUnsignedByteAsInt();

        switch (_IntervallKurz_) {
            case 1:
            case 2:
            case 4:
            case 8:
            case 12:
            case 16:
            case 20:
            case 24:
            case 40:
            case 48:
            case 60:
            case 80:
            case 120:
            case 240:
                data.getUnscaledValue("IntervallDauerKurzZeitDaten").set(_IntervallKurz_);
                _intervalTime = _IntervallKurz_ * 15;// Angaben sind in 15-Sekunden Intervallen
                break;
            default:
                debug.warning("IntervallDauerKurzZeitDaten ungültig : " + _IntervallKurz_ + Debug.NEWLINE + "De-Block (Osi7Adresse-DeNummer-FG-Typ-Id): " + deBlock.getLongKey() + " verworfen:" + Debug.NEWLINE + deBlock.toString());
                return resultData;
        }

        // VersionLang-Byte aus Telegramm ------------------------------------------------------------------------------
        int _VersionLang_ = deBlock.getDeData().readUnsignedByteAsInt();
        switch (_VersionLang_) {
            case 10:
            case 11:
            case 12:
            case 13:
            case 14:
            case 15:
            case 16:
            case 17:
            case 18:
            case 19:
            case 20:
            case 21:
            case 22:
            case 23:
            case 255:
                data.getUnscaledValue("VersionLangZeitDaten").set(_VersionLang_);
                break;
            default:
                debug.warning("VersionLangZeitDaten ungültig : " + _VersionLang_ + Debug.NEWLINE + "De-Block (Osi7Adresse-DeNummer-FG-Typ-Id): " + deBlock.getLongKey() + " verworfen:" + Debug.NEWLINE + deBlock.toString());
                return resultData;
        }

        // Întervalldaten Langzeit-Byte aus Telegramm ------------------------------------------------------------------
        int _IntervallLang_ = deBlock.getDeData().readUnsignedByteAsInt();

        switch (_IntervallLang_) {
            case 129:
            case 130:
            case 131:
            case 132:
            case 134:
            case 136:
            case 140:
            case 152:
                data.getUnscaledValue("IntervallDauerLangZeit").set(_IntervallLang_);
                break;
            default:
                // Wenn Langzeit gar nicht aktiv, falsche Zeitdauer ignorieren und auf 1h setzen.
                if (_VersionLang_ == 255) {
                    data.getUnscaledValue("IntervallDauerLangZeit").set(129);
                }
                else {
                    debug.warning("IntervallDauerLangZeit ungültig : " + _IntervallLang_ + Debug.NEWLINE + "De-Block (Osi7Adresse-DeNummer-FG-Typ-Id): " + deBlock.getLongKey() + " verworfen:" + Debug.NEWLINE + deBlock.toString());
                    return resultData;
                }
        }

        // Alpha 1-Byte aus Telegramm ----------------------------------------------------------------------------------
        int _alpha1_ = deBlock.getDeData().readUnsignedByteAsInt();
        data.getUnscaledValue("alpha1").set(_alpha1_ + 1);

        // Alpha 2-Byte aus Telegramm ----------------------------------------------------------------------------------
        int _alpha2_ = deBlock.getDeData().readUnsignedByteAsInt();
        data.getUnscaledValue("alpha2").set(_alpha2_ + 1);

        // Längengrenzwert für LKW-Byte aus Telegramm ------------------------------------------------------------------
        int _LaengeLKW_ = deBlock.getDeData().readUnsignedByteAsInt();
        if (_LaengeLKW_ == 255) {
            data.getUnscaledValue("LängenGrenzWert").setText("kein Längengrenzwert");
        }
        else {
            data.getUnscaledValue("LängenGrenzWert").set(_LaengeLKW_ + 400);
        }

        // Startwert für Mittelwertbilung-Byte aus Telegramm -----------------------------------------------------------
        int _MittelWert_ = deBlock.getDeData().readUnsignedByteAsInt();

        int _MittelWertBildung_ = _MittelWert_ & 0x80;
        int _MittelWertStart_ = _MittelWert_ & 0x7f;

        if (_MittelWertBildung_ == 0) {
            data.getUnscaledValue("ArtMittelWertBildung").setText("gleitende Mittelwertbildung");
            _kindOfAverage = 0;
        }
        else {
            data.getUnscaledValue("ArtMittelWertBildung").setText("arithmetische Mittelwertbildung");
            _kindOfAverage = 1;
        }

        data.getUnscaledValue("StartMittelWertBildung").set(_MittelWertStart_);

        //Ergebnisdatensatz erzeugen...
        resultData = new ResultData(_de.getDe(), dataDescription, _eakConversionModul.getTimeStamp(), data);
        return resultData;
    }

    /**
     * FG 1 : Typ: 128 Id: 4 Name: DaLVEAnalyseWerteOpt0 am MessQuerschnitt<br>
     * <p>
     * Beschreibung des Telegrammaufbaus und Inhalts siehe db_desc und snd-LEV-DAV von Heusch/Boesefeldt.
     * <pre>
     *
     * Fg   1 Id 132 Typ 128 ------------------------------ Analysewerte Option 0
     * 	 Datatype DaLVEAnalyseWerteOpt0          mit Zeit(UnixZeit,
     * UnixArchivZeit)
     * 	 Reaktion       Write Area
     * 	 NUR_AKTUELL_SENDEN
     * 	 Fill High     (QKFZ)           Stundenwert KFZ-Menge in KFZ/h
     * 	 Fill Low      (QKFZ)
     *
     * 	 Fill High     (QLKW)           Stundenwert LKW-Menge in LKW/h
     * 	 Fill Low      (QLKW)
     *
     * 	 Fill High     (QPKW)           Stundenwert PKW-Menge in PKW/h
     * 	 Fill Low      (QPKW)
     *
     * 	 Fill Low      (VKFZ)           mittlere Geschwindigkeit KFZ in km/h
     * 	 Fill Low      (VLKW)           mittlere Geschwindigkeit LKW in km/h
     * 	 Fill Low      (VPKW)           mittlere Geschwindigkeit PKW in km/h
     * 	 Fill Low      (aLKW)           Anteil LKW in %
     *
     * 	 Fill High     (K)              K = QKFZ/VKFZ in KFZ/km
     * 	 Fill Low      (K)
     * </pre>
     *
     * @param deBlock         Der zu bearbeitende De-Block.
     * @param data            Der passende Datensatz für die Ziel Attributgruppe/Aspekt-Kombination.
     * @param dataDescription Die Datenbeschreibung für die Attributgruppe/Aspekt-Kombination.
     *
     * @return Ergebnisdatensatz mit den vom DeBlock in die entsprechende Attributgruppe konvertierten Daten.
     */
    public ResultData Fg001Typ128Id004AtgVerkehrsDatenKurzZeitMqAspAnalyseHB(DeBlock deBlock, Data data, DataDescription dataDescription) {
        ResultData resultData = null;

        Data att;
        Data status;
        Data erfassung;
        Data plFormal;
        Data plLogisch;
        Data ersetzung;
        Data güte;

        String[] Attribute = {"QKfz", "VKfz", "QLkw", "VLkw", "QPkw", "VPkw", "B", "BMax", "SKfz", "VgKfz", "ALkw", "KKfz", "KLkw", "KPkw", "QB", "KB", "VDelta"};

        // Initialisierung auf Standardwerte
        for (int i = 0; i < Attribute.length; ++i) {
            att = data.getItem(Attribute[i]);
            status = att.getItem("Status");
            erfassung = status.getItem("Erfassung");
            plFormal = status.getItem("PlFormal");
            plLogisch = status.getItem("PlLogisch");
            ersetzung = status.getItem("MessWertErsetzung");
            güte = att.getItem("Güte");

            att.getTextValue("Wert").setText("nicht ermittelbar");

            erfassung.getTextValue("NichtErfasst").setText("Nein");

            plFormal.getTextValue("WertMax").setText("Nein");
            plFormal.getTextValue("WertMin").setText("Nein");

            plLogisch.getTextValue("WertMaxLogisch").setText("Nein");
            plLogisch.getTextValue("WertMinLogisch").setText("Nein");

            ersetzung.getTextValue("Implausibel").setText("Nein");
            ersetzung.getTextValue("Interpoliert").setText("Nein");

            güte.getScaledValue("Index").set(1.0);// Keine Güteinfo von HB, auf 1.0 setzen
            güte.getTextValue("Verfahren").setText("Standard");
        }

        // Werte aus Telegramm -----------------------------------------------------------------------------------------
        int _QKFZ_ = deBlock.getDeData().readUnsignedHighLowByteAsInt();
        int _QLKW_ = deBlock.getDeData().readUnsignedHighLowByteAsInt();
        int _QPKW_ = deBlock.getDeData().readUnsignedHighLowByteAsInt();
        int _VKFZ_ = deBlock.getDeData().readUnsignedByteAsInt();
        int _VLKW_ = deBlock.getDeData().readUnsignedByteAsInt();
        int _VPKW_ = deBlock.getDeData().readUnsignedByteAsInt();
        int _aLKW_ = deBlock.getDeData().readUnsignedByteAsInt();
        int _KKFZ_ = deBlock.getDeData().readUnsignedHighLowByteAsInt();

        // Setze QKfz
        if (tls16BitToInt(_QKFZ_) != -1) {
            data.getItem("QKfz").getUnscaledValue("Wert").set(_QKFZ_);
        }
        // Setze VKfz (Wenn Wert == -1 (nicht ermittelbar) oder 0 km/h, dann "nicht ermittelbar" übertragen.
        if (tls8BitToInt(_VKFZ_) > 0) {
            data.getItem("VKfz").getUnscaledValue("Wert").set(_VKFZ_);
        }
        // Setze QLkw
        if (tls16BitToInt(_QLKW_) != -1) {
            data.getItem("QLkw").getUnscaledValue("Wert").set(_QLKW_);
        }
        // Setze VLkw (Wenn Wert == -1 (nicht ermittelbar) oder 0 km/h, dann "nicht ermittelbar" übertragen.
        if (tls8BitToInt(_VLKW_) > 0) {
            data.getItem("VLkw").getUnscaledValue("Wert").set(_VLKW_);
        }
        // Setze QPkw
        if (tls16BitToInt(_QPKW_) != -1) {
            data.getItem("QPkw").getUnscaledValue("Wert").set(_QPKW_);
        }
        // Setze VPkw (Wenn Wert == -1 (nicht ermittelbar) oder 0 km/h, dann "nicht ermittelbar" übertragen.
        if (tls8BitToInt(_VPKW_) > 0) {
            data.getItem("VPkw").getUnscaledValue("Wert").set(_VPKW_);
        }
        // Setze ALkw
        if (tls8BitToInt(_aLKW_) != -1) {
            data.getItem("ALkw").getUnscaledValue("Wert").set(_aLKW_);
        }
        // Setze KKfz (Wenn Wert == -1 (nicht ermittelbar) oder 0 km/h, dann "nicht ermittelbar" übertragen.
        if (tls16BitToInt(_KKFZ_) > 0) {
            data.getItem("KKfz").getUnscaledValue("Wert").set(_KKFZ_);
        }

        //Ergebnisdatensatz erzeugen...
        resultData = new ResultData(_de.getDe(), dataDescription, _eakConversionModul.getTimeStamp(), data);
        return resultData;
    }

    /**
     * Prüft, ob der übergebene Wert == 0xff ist. In diesem Fall wird -1 zurückgegeben, so dass dieser als unscaledValue
     * direkt gesetzt werden kann (entspricht dem Zustand "nicht ermittelbar").
     *
     * @param value Der zu prüfende Wert (der als 8 Bit vorzeichenlos interpretiert wird)
     *
     * @return value, wenn ungleich 255, sonst -1.
     */
    private int tls8BitToInt(int value) {
        if (value == 0xff) {
            return -1;
        }
        else {
            return value;
        }
    }

    /**
     * Prüft, ob der übergebene Wert == 0xffff ist. In diesem Fall wird -1 zurückgegeben, so dass dieser als
     * unscaledValue direkt gesetzt werden kann (entspricht dem Zustand "nicht ermittelbar").
     *
     * @param value Der zu prüfende Wert (der als 16 Bit vorzeichenlos interpretiert wird)
     *
     * @return value, wenn ungleich 0xffff, sonst -1.
     */
    private int tls16BitToInt(int value) {
        if (value == 0xffff) {
            return -1;
        }
        else {
            return value;
        }
    }

    /**
     * FG 1 : Typ: 144 Id: 4 Name: DaLVEVLageSpOpt0 am FahrStreifen<br>
     * <p>
     * Beschreibung des Telegrammaufbaus und Inhalts siehe db_desc und snd-LEV-DAV von Heusch/Boesefeldt.
     *
     * @param deBlock         Der zu bearbeitende De-Block.
     * @param data            Der passende Datensatz für die Ziel Attributgruppe/Aspekt-Kombination.
     * @param dataDescription Die Datenbeschreibung für die Attributgruppe/Aspekt-Kombination.
     *
     * @return Ergebnisdatensatz mit den vom DeBlock in die entsprechende Attributgruppe konvertierten Daten.
     */
    public ResultData Fg001Typ144Id004AtgStörfallZustandAspStörfallVerfahrenNRWHB(DeBlock deBlock, Data data, DataDescription dataDescription) {
        ResultData resultData = null;

        // VLage-Byte aus Telegramm ----------------------------------------------------------------------------
        int _VLage_ = deBlock.getDeData().readUnsignedByteAsInt();

        data.getTimeValue("T").setSeconds(60);

        data.getUnscaledValue("Situation").set(_VLage_);

        data.getTimeValue("Horizont").setSeconds(0);// Analysewerte, kein Prognosehorizont
        data.getItem("Güte").getScaledValue("Index").set(1.0);// Keine Güteinfo von HB, auf 1.0 setzen
        data.getItem("Güte").getTextValue("Verfahren").setText("Standard");

        //Ergebnisdatensatz erzeugen...
        resultData = new ResultData(_de.getDe(), dataDescription, _eakConversionModul.getTimeStamp(), data);
        return resultData;
    }

    /**
     * FG 1 : Typ: 152 Id: 4 Name: DaLVEVLageOpt0 am MessQuerschnitt<br>
     * <p>
     * Beschreibung des Telegrammaufbaus und Inhalts siehe db_desc und snd-LEV-DAV von Heusch/Boesefeldt.
     *
     * @param deBlock         Der zu bearbeitende De-Block.
     * @param data            Der passende Datensatz für die Ziel Attributgruppe/Aspekt-Kombination.
     * @param dataDescription Die Datenbeschreibung für die Attributgruppe/Aspekt-Kombination.
     *
     * @return Ergebnisdatensatz mit den vom DeBlock in die entsprechende Attributgruppe konvertierten Daten.
     */
    public ResultData Fg001Typ152Id004AtgStörfallZustandAspStörfallVerfahrenNRWHB(DeBlock deBlock, Data data, DataDescription dataDescription) {
        ResultData resultData = null;

        // VLage-Byte aus Telegramm ----------------------------------------------------------------------------
        int _VLage_ = deBlock.getDeData().readUnsignedByteAsInt();

        data.getTimeValue("T").setSeconds(60);

        data.getUnscaledValue("Situation").set(_VLage_);

        data.getTimeValue("Horizont").setSeconds(0);// Analysewerte, kein Prognosehorizont
        data.getItem("Güte").getScaledValue("Index").set(1.0);// Keine Güteinfo von HB, auf 1.0 setzen
        data.getItem("Güte").getTextValue("Verfahren").setText("Standard");

        //Ergebnisdatensatz erzeugen...
        resultData = new ResultData(_de.getDe(), dataDescription, _eakConversionModul.getTimeStamp(), data);
        return resultData;
    }

    public ResultData Fg001Typ160Id004AtgTlsLveErgebnisMeldungVersion0Bis4AspTlsAntwort(DeBlock deBlock, Data data, DataDescription dataDescription) {
        return TlsLveErgebnisMeldungVersion0Bis4AspTlsAntwort(deBlock, data, dataDescription);
    }

    /**
     * Verarbeitet alls Telegrammversionen der TLS Kurzeit Version 0-4 (Typ 49/113 -53/117).
     *
     * @param deBlock         Der zu bearbeitende De-Block.
     * @param data            Der passende Datensatz für die Ziel Attributgruppe/Aspekt-Kombination.
     * @param dataDescription Die Datenbeschreibung für die Attributgruppe/Aspekt-Kombination.
     *
     * @return Ergebnisdatensatz mit den vom DeBlock in die entsprechende Attributgruppe konvertierten Daten.
     */
    private ResultData TlsLveErgebnisMeldungVersion0Bis4AspTlsAntwort(DeBlock deBlock, Data data, DataDescription dataDescription) {
        ResultData resultData = null;

        // Intervalllänge aus IntervalldatenHeader ---------------------------------------------------------------------
        long intervallTime = _eakConversionModul.getLveIntervalDuration();
        if (intervallTime < 0) {
            //Ungültige Intervallzeit (HB-spezial), dann auf 60 Sekunden einstellen
            data.getTimeValue("T").setSeconds(getIntervallTime());
        }
        else {
            data.getTimeValue("T").setSeconds(intervallTime);
        }

        // Werte jeweils auf Zustand "nicht ermittelbar" setzen.
        data.getUnscaledValue("qKfz").setText("nicht ermittelbar");
        data.getUnscaledValue("qLkwÄ").setText("nicht ermittelbar");
        data.getUnscaledValue("vPkwÄ").setText("nicht ermittelbar");
        data.getUnscaledValue("vLkwÄ").setText("nicht ermittelbar");
        data.getUnscaledValue("tNetto").setText("nicht ermittelbar");
        data.getUnscaledValue("b").setText("nicht ermittelbar");
        data.getUnscaledValue("sKfz").setText("nicht ermittelbar");
        data.getUnscaledValue("vKfz").setText("nicht ermittelbar");
        data.getUnscaledArray("qPkwÄGeschwKlasse").setLength(0);
        data.getUnscaledArray("qLkwÄGeschwKlasse").setLength(0);

        // Werte aus Telegramm -----------------------------------------------------------------------------------------

        int _qKfz_;// qKfz-Wert aus Telegramm
        int _qLkwÄ_;// qLkw-Wert aus Telegramm
        int _vPkwÄ_;// qPkwÄ-Wert aus Telegramm
        int _vLkwÄ_;// vLkwÄ-Wert aus Telegramm
        int _tNetto_;// tNetto-Wert aus Telegramm
        int _b_;// Belegungs-Wert aus Telegramm
        int _sKfz_;// Streuung Kfz-Wert aus Telegramm
        int _vKfz_;// vKfz-Wert aus Telegramm

        // Telegrammabhängig die Daten einlesen, im Telegramm nicht vorhandene Daten auf "nicht ermittelbar" lassen

        _qKfz_ = deBlock.getDeData().readUnsignedHighLowByteAsInt();
        data.getUnscaledValue("qKfz").set(_qKfz_);
        _qLkwÄ_ = deBlock.getDeData().readUnsignedHighLowByteAsInt();
        data.getUnscaledValue("qLkwÄ").set(_qLkwÄ_);

        _vPkwÄ_ = deBlock.getDeData().readUnsignedByteAsInt();
        if (_vPkwÄ_ != 0) {
            data.getUnscaledValue("vPkwÄ").set(_vPkwÄ_);
        }
        else {
            data.getUnscaledValue("vPkwÄ").setText("nicht ermittelbar");
        }

        _vLkwÄ_ = deBlock.getDeData().readUnsignedByteAsInt();
        if (_vLkwÄ_ != 0) {
            data.getUnscaledValue("vLkwÄ").set(_vLkwÄ_);
        }
        else {
            data.getUnscaledValue("vLkwÄ").setText("nicht ermittelbar");
        }

        _tNetto_ = deBlock.getDeData().readUnsignedByteAsInt();
        data.getUnscaledValue("tNetto").set(_tNetto_);
        _b_ = deBlock.getDeData().readUnsignedByteAsInt();
        if (_b_ != 255) {
            if (_b_ > 100) {
                debug.warning("Belegung von " + _b_ + "% > 100% erhalten! Wert wurde auf 'nicht ermittelbar' geändert!");
            }
            else {
                data.getUnscaledValue("b").set(_b_);
            }
        }
        _sKfz_ = deBlock.getDeData().readUnsignedByteAsInt();
        data.getUnscaledValue("sKfz").set(_sKfz_);
        _vKfz_ = deBlock.getDeData().readUnsignedByteAsInt();
        data.getUnscaledValue("vKfz").set(_vKfz_);

        //Ergebnisdatensatz erzeugen...
        resultData = new ResultData(_de.getDe(), dataDescription, _eakConversionModul.getTimeStamp(), data);
        return resultData;
    }

    /**
     * FG 1 : Typ: 160 Id: 4 Name: DaLVEKurzZeitInterpolOpt0 am FahrStreifen<br>
     * <p>
     * Beschreibung des Telegrammaufbaus und Inhalts siehe db_desc und snd-LEV-DAV von Heusch/Boesefeldt.
     * <pre>
     *   Fg   1 Id 132 Typ 160 --------------- Intervalldaten Kurzzeit
     * Interpoliert
     *   Datatype DaLVEKurzInterpolOpt0          mit Zeit (UnixZeit,
     * UnixArchivZeit)
     *           Reaktion       Write Area
     *    NUR_AKTUELL_SENDEN
     *    Fill High     (qKFZ)           KFZ-Menge
     *    Fill Low      (qKFZ)
     *
     *    Fill High     (qLKW)           LKW-Menge
     *    Fill Low      (qLKW)
     *
     *    Fill Low      (vPKW)           PKW-Geschwindigkeit
     *    Fill Low      (vLKW)           LKW-Geschwindigkeit
     *    Fill Low      (tNetto)         Nettozeitluecke in Sekunden (0..254)
     *    Fill Low      (Beleg)          Belegungsgrad in 0..100%
     *    Fill Low      (s)              Streung in km/h
     *
     *  #ifndef OHNE_VKFZ_MITTEL
     *    Fill Low      (vKFZMittel)     vExpo gemaes TLS Option 3 in km/h
     *  #endif
     *
     * </pre>
     *
     * @param deBlock         Der zu bearbeitende De-Block.
     * @param data            Der passende Datensatz für die Ziel Attributgruppe/Aspekt-Kombination.
     * @param dataDescription Die Datenbeschreibung für die Attributgruppe/Aspekt-Kombination.
     *
     * @return Ergebnisdatensatz mit den vom DeBlock in die entsprechende Attributgruppe konvertierten Daten.
     */
    public ResultData Fg001Typ160Id004AtgVerkehrsDatenKurzZeitIntervallAspMessWertErsetzung(DeBlock deBlock, Data data, DataDescription dataDescription) {
        ResultData resultData = null;

        Data att;
        Data status;
        Data erfassung;
        Data plFormal;
        Data plLogisch;
        Data ersetzung;
        Data güte;

        String[] Attribute = {"qKfz", "vKfz", "qLkw", "vLkw", "qPkw", "vPkw", "b", "tNetto", "sKfz", "vgKfz"};

        // Initialisierung auf Standardwerte
        for (int i = 0; i < Attribute.length; ++i) {
            att = data.getItem(Attribute[i]);
            status = att.getItem("Status");
            erfassung = status.getItem("Erfassung");
            plFormal = status.getItem("PlFormal");
            plLogisch = status.getItem("PlLogisch");
            ersetzung = status.getItem("MessWertErsetzung");
            güte = att.getItem("Güte");

            att.getTextValue("Wert").setText("nicht ermittelbar");

            erfassung.getTextValue("NichtErfasst").setText("Nein");

            plFormal.getTextValue("WertMax").setText("Nein");
            plFormal.getTextValue("WertMin").setText("Nein");

            plLogisch.getTextValue("WertMaxLogisch").setText("Nein");
            plLogisch.getTextValue("WertMinLogisch").setText("Nein");

            ersetzung.getTextValue("Implausibel").setText("Nein");
            ersetzung.getTextValue("Interpoliert").setText("Nein");

            güte.getScaledValue("Index").set(1.0);// Keine Güteinfo von HB, auf 1.0 setzen
            güte.getTextValue("Verfahren").setText("Standard");
        }

        // Werte aus Telegramm -----------------------------------------------------------------------------------------

        int _qKFZ_ = deBlock.getDeData().readUnsignedHighLowByteAsInt();
        int _qLKW_ = deBlock.getDeData().readUnsignedHighLowByteAsInt();
        int _vPKW_ = deBlock.getDeData().readUnsignedByteAsInt();
        int _vLKW_ = deBlock.getDeData().readUnsignedByteAsInt();
        int _tNetto_ = deBlock.getDeData().readUnsignedByteAsInt() * 100;// Umrechnung in Millisekunden
        int _b_ = deBlock.getDeData().readUnsignedByteAsInt();
        int _s_ = deBlock.getDeData().readUnsignedByteAsInt();
        int _vKfzMittel_ = deBlock.getDeData().readUnsignedByteAsInt();

        int _qPKW_ = -1;
        if ((_qKFZ_ != 0xffff) && (_qLKW_ != 0xffff)) {
            _qPKW_ = (_qKFZ_ - _qLKW_) > 0 ? (_qKFZ_ - _qLKW_) : 0;
        }

        int _vKFZ_ = -1;
        if ((_qLKW_ != 0xffff) && (_qPKW_ != 0xffff) && (_vLKW_ != 0xff) && (_vPKW_ != 0xff)) // LKW und PKW mit V
        {
            _vKFZ_ = (_qLKW_ + _qPKW_) > 0 ? ((_qLKW_ * _vLKW_) + (_qPKW_ * _vPKW_)) / (_qLKW_ + _qPKW_) : -1;
        }
        else if ((_qLKW_ != 0xffff) && (_qLKW_ > 0) && (_qPKW_ == 0) && (_vLKW_ != 0xff)) // nur LKW mit V
        {
            _vKFZ_ = _vLKW_;
        }
        else if ((_qPKW_ != 0xffff) && (_qPKW_ > 0) && (_qLKW_ == 0) && (_vPKW_ != 0xff)) // nur PKW mit V
        {
            _vKFZ_ = _vPKW_;
        }

        // Intervalllänge aus IntervalldatenHeader ---------------------------------------------------------------------
        long intervallTime = _eakConversionModul.getLveIntervalDuration();
        if (intervallTime < 0) {
            //debug.error("HB IntervallTime: " + intervallTime);
            data.getTimeValue("T").setSeconds(getIntervallTime());
        }
        else {
            data.getTimeValue("T").setSeconds(intervallTime);
        }

        // Art der Mittelwertbildung am DE aus Betriebsparametern ------------------------------------------------------
        switch (_kindOfAverage) {
            case 0:
                data.getUnscaledValue("ArtMittelwertbildung").setText("gleitende Mittelwertbildung");
                break;
            case 1:
                data.getUnscaledValue("ArtMittelwertbildung").setText("arithmetische Mittelwertbildung");
                break;
            default:
                data.getUnscaledValue("ArtMittelwertbildung").setText("unbekannt");
        }

        data.getUnscaledValue("ArtMittelwertbildung").set(_kindOfAverage);
        // Setze QKfz
        if (tls16BitToInt(_qKFZ_) != -1) {
            data.getItem("qKfz").getUnscaledValue("Wert").set(_qKFZ_);
        }
        // Setze VKfz (Wenn Wert == -1 (nicht ermittelbar) oder 0 km/h, dann "nicht ermittelbar" übertragen.
        if (tls8BitToInt(_vKFZ_) > 0) {
            data.getItem("vKfz").getUnscaledValue("Wert").set(_vKFZ_);
        }
        // Setze QLkw
        if (tls16BitToInt(_qLKW_) != -1) {
            data.getItem("qLkw").getUnscaledValue("Wert").set(_qLKW_);
        }
        // Setze VLkw (Wenn Wert == -1 (nicht ermittelbar) oder 0 km/h, dann "nicht ermittelbar" übertragen.
        if (tls8BitToInt(_vLKW_) > 0) {
            data.getItem("vLkw").getUnscaledValue("Wert").set(_vLKW_);
        }
        // Setze QPkw
        if (tls16BitToInt(_qPKW_) != -1) {
            data.getItem("qPkw").getUnscaledValue("Wert").set(_qPKW_);
        }
        // Setze VPkw (Wenn Wert == -1 (nicht ermittelbar) oder 0 km/h, dann "nicht ermittelbar" übertragen.
        if (tls8BitToInt(_vPKW_) > 0) {
            data.getItem("vPkw").getUnscaledValue("Wert").set(_vPKW_);
        }
        // Setze Belegung
        if (tls8BitToInt(_b_) != -1) {
            data.getItem("b").getUnscaledValue("Wert").set(_b_);
        }
        // Setze Nettozeitlücke
        if (tls8BitToInt(_tNetto_) != -1) {
            data.getItem("tNetto").getUnscaledValue("Wert").set(_tNetto_);
        }
        // Setze Streueung Kfz
        if (tls8BitToInt(_s_) != -1) {
            data.getItem("sKfz").getUnscaledValue("Wert").set(_s_);
        }
        // Setze vKfzMittel
        if (tls8BitToInt(_vKfzMittel_) != -1) {
            data.getItem("vgKfz").getUnscaledValue("Wert").set(_vKfzMittel_);
        }

        //Ergebnisdatensatz erzeugen...
        resultData = new ResultData(_de.getDe(), dataDescription, _eakConversionModul.getTimeStamp(), data);
        return resultData;
    }

    /**
     * FG 1 : Typ: 182 Id: 4 Name: DaLVEVLageAbschnitt am StraßenSegment<br>
     * <p>
     * Beschreibung des Telegrammaufbaus und Inhalts siehe db_desc und snd-LEV-DAV von Heusch/Boesefeldt.
     *
     * @param deBlock         Der zu bearbeitende De-Block.
     * @param data            Der passende Datensatz für die Ziel Attributgruppe/Aspekt-Kombination.
     * @param dataDescription Die Datenbeschreibung für die Attributgruppe/Aspekt-Kombination.
     *
     * @return Ergebnisdatensatz mit den vom DeBlock in die entsprechende Attributgruppe konvertierten Daten.
     */
    public ResultData Fg001Typ182Id004AtgStörfallZustandAspStörfallVerfahrenRDSHB(DeBlock deBlock, Data data, DataDescription dataDescription) {
        ResultData resultData = null;

        // VLage-Byte aus Telegramm ----------------------------------------------------------------------------
        int _VLage_ = deBlock.getDeData().readUnsignedByteAsInt();

        data.getTimeValue("T").setSeconds(60);

        data.getUnscaledValue("Situation").set(_VLage_);

        data.getTimeValue("Horizont").setSeconds(0);// Analysewerte, kein Prognosehorizont
        data.getItem("Güte").getScaledValue("Index").set(1.0);// Keine Güteinfo von HB, auf 1.0 setzen
        data.getItem("Güte").getTextValue("Verfahren").setText("Standard");

        //Ergebnisdatensatz erzeugen...
        resultData = new ResultData(_de.getDe(), dataDescription, _eakConversionModul.getTimeStamp(), data);
        return resultData;
    }

    /**
     * Verarbeitet Langzeittelegramm der VRZ Leverkusen (Datatype DaLVELang2000
     *
     * @param deBlock         Der zu bearbeitende De-Block.
     * @param data            Der passende Datensatz für die Ziel Attributgruppe/Aspekt-Kombination.
     * @param dataDescription Die Datenbeschreibung für die Attributgruppe/Aspekt-Kombination.
     *
     * @return Ergebnisdatensatz mit den vom DeBlock in die entsprechende Attributgruppe konvertierten Daten.
     */
    public ResultData Fg001Typ204Id004AtgVerkehrsDatenLangZeitIntervallAspExterneErfassung(DeBlock deBlock, Data data, DataDescription dataDescription) {
        ResultData resultData = null;

        int _qKfz_;
        int _qPkwÄ_;// für Erweiterungen aus Gründen der Systematik mit aufgenommen.
        int _qKfzNk_;
        int _qPkwG_;
        int _qPkw_;
        int _qKrad_;
        int _qLfw_;
        int _qLkwÄ_;
        int _qPkwA_;
        int _qLkw_;
        int _qBus_;
        int _qLkwK_;
        int _qLkwA_;
        int _qSattelKfz_;
        int _vKfz_;
        int _vPkwÄ_;
        int _vKfzNk_;// für Erweiterungen aus Gründen der Systematik mit aufgenommen.
        int _vPkwG_;
        int _vPkw_;
        int _vKrad_;
        int _vLfw_;
        int _vLkwÄ_;
        int _vPkwA_;
        int _vLkw_;
        int _vBus_;
        int _vLkwK_;
        int _vLkwA_;
        int _vSattelKfz_;
        int _sKfz_;// für Erweiterungen aus Gründen der Systematik mit aufgenommen.
        int _sPkwÄ_;
        int _sKfzNk_;// für Erweiterungen aus Gründen der Systematik mit aufgenommen.
        int _sPkwG_;
        int _sPkw_;
        int _sKrad_;
        int _sLfw_;
        int _sLkwÄ_;
        int _sPkwA_;
        int _sLkw_;
        int _sBus_;
        int _sLkwK_;
        int _sLkwA_;// für Erweiterungen aus Gründen der Systematik mit aufgenommen.
        int _sSattelKfz_;
        int _v85Kfz_;// für Erweiterungen aus Gründen der Systematik mit aufgenommen.
        int _v85PkwÄ_;
        int _v85KfzNk_;// für Erweiterungen aus Gründen der Systematik mit aufgenommen.
        int _v85PkwG_;// für Erweiterungen aus Gründen der Systematik mit aufgenommen.
        int _v85Pkw_;// für Erweiterungen aus Gründen der Systematik mit aufgenommen.
        int _v85Krad_;// für Erweiterungen aus Gründen der Systematik mit aufgenommen.
        int _v85Lfw_;// für Erweiterungen aus Gründen der Systematik mit aufgenommen.
        int _v85LkwÄ_;// für Erweiterungen aus Gründen der Systematik mit aufgenommen.
        int _v85PkwA_;// für Erweiterungen aus Gründen der Systematik mit aufgenommen.
        int _v85Lkw_;// für Erweiterungen aus Gründen der Systematik mit aufgenommen.
        int _v85Bus_;// für Erweiterungen aus Gründen der Systematik mit aufgenommen.
        int _v85LkwK_;// für Erweiterungen aus Gründen der Systematik mit aufgenommen.
        int _v85LkwA_;// für Erweiterungen aus Gründen der Systematik mit aufgenommen.
        int _v85SattelKfz_;// für Erweiterungen aus Gründen der Systematik mit aufgenommen.

        // Werte jeweils auf Zustand "nicht ermittelbar" setzen.
        int count = 0;
        int _qKlasse_;

        Data.Array array = null;

        // Werte jeweils auf Zustand "nicht ermittelbar" setzen.

        String[] valStrings = {"qKfz", "qPkwÄ", "qKfzNk", "qPkwG", "qPkw", "qKrad", "qLfw", "qLkwÄ", "qPkwA", "qLkw", "qBus", "qLkwK", "qLkwA", "qSattelKfz", "vKfz", "vPkwÄ", "vKfzNk", "vPkwG", "vPkw", "vKrad", "vLfw", "vLkwÄ", "vPkwA", "vLkw", "vBus", "vLkwK", "vLkwA", "vSattelKfz", "sKfz", "sPkwÄ", "sKfzNk", "sPkwG", "sPkw", "sKrad", "sLfw", "sLkwÄ", "sPkwA", "sLkw", "sBus", "sLkwK", "sLkwA", "sSattelKfz", "v85Kfz", "v85PkwÄ", "v85KfzNk", "v85PkwG", "v85Pkw", "v85Krad", "v85Lfw", "v85LkwÄ", "v85PkwA", "v85Lkw", "v85Bus", "v85LkwK", "v85LkwA", "v85SattelKfz"};

        for (String valString : valStrings) {
            data.getItem(valString).getUnscaledValue("Wert").setText("nicht ermittelbar");
            data.getItem(valString).getItem("Status").getItem("Erfassung").getUnscaledValue("NichtErfasst").setText("Nein");
            data.getItem(valString).getItem("Status").getItem("PlFormal").getUnscaledValue("WertMax").setText("Nein");
            data.getItem(valString).getItem("Status").getItem("PlFormal").getUnscaledValue("WertMin").setText("Nein");
            data.getItem(valString).getItem("Status").getItem("PlLogisch").getUnscaledValue("WertMaxLogisch").setText("Nein");
            data.getItem(valString).getItem("Status").getItem("PlLogisch").getUnscaledValue("WertMinLogisch").setText("Nein");
            data.getItem(valString).getItem("Status").getItem("MessWertErsetzung").getUnscaledValue("Implausibel").setText("Nein");
            data.getItem(valString).getItem("Status").getItem("MessWertErsetzung").getUnscaledValue("Interpoliert").setText("Nein");
            data.getItem(valString).getItem("Güte").getScaledValue("Index").set(1.0);// Keine Güteinfo von HB, auf 1.0 setzen
            data.getItem(valString).getItem("Güte").getUnscaledValue("Verfahren").set(0);
        }

        data.getArray("qKfzGeschwKlasse").setLength(0);
        data.getArray("qPkwÄGeschwKlasse").setLength(0);
        data.getArray("qKfzNkGeschwKlasse").setLength(0);
        data.getArray("qPkwGGeschwKlasse").setLength(0);
        data.getArray("qPkwGeschwKlasse").setLength(0);
        data.getArray("qKradGeschwKlasse").setLength(0);
        data.getArray("qLfwGeschwKlasse").setLength(0);
        data.getArray("qLkwÄGeschwKlasse").setLength(0);
        data.getArray("qPkwAGeschwKlasse").setLength(0);
        data.getArray("qLkwGeschwKlasse").setLength(0);
        data.getArray("qBusGeschwKlasse").setLength(0);
        data.getArray("qLkwKGeschwKlasse").setLength(0);
        data.getArray("qLkwAGeschwKlasse").setLength(0);
        data.getArray("qSattelKfzGeschwKlasse").setLength(0);

        // Intervallbeginn als UNIX-Zeit einlesen
        long _UnixZeit_ = deBlock.getDeData().readUnsignedVeryHighExtraHighHighLowByteAsInt();

        // Zeitstempel kommt häufig fehlerhaft !! als 0 zurück ,
        // in der Regel sind dann auch alle anderen Werte 0
        long offsetHBtoUTC = 0;// Zeitdifferenz in Sekunden der HB-UNIX-Zeit zu UTC.
        if (_UnixZeit_ > 3600) {
            offsetHBtoUTC = -3600; // Zeitdifferenz in Sekunden der HB-UNIX-Zeit zu UTC.
            // Die vorstehnde Zeile war bis zum 02.10.2020 (wahrscheinlich seit 2008) mit folgendem Kommentar ausgeschaltet:
	        // "Augenscheinlich gibts diese Abweichung nicht bei Langzeitdaten, sonst schon. Komisch!"
	        //
	        // Seit dem 02.10.2020 ist die Zeitkorrektur wieder aktiv (ab KExTlsOsi7 V5.8.0), weil die Differenz in der Praxis
	        // jetzt wieder vorliegt.
        }
        _UnixZeit_ = (((_UnixZeit_ + offsetHBtoUTC) * 1000));

        // Intervalllänge einlesen: Wert = Intervalllänge in Stunden (KEINE TLS Kodierung)
        long _intervallLaenge_ = deBlock.getDeData().readUnsignedByteAsInt() * 60 * 60;// in s
        data.getTimeValue("T").setSeconds(_intervallLaenge_);

        // Versionsangabe überlesen
        deBlock.getDeData().readUnsignedByteAsInt();

        // AnzahlFahrzeugKlassenCodes überlesen
        deBlock.getDeData().readUnsignedByteAsInt();

        // Fahrzeugklassen festlegen (gemäß TLS)
        Map<Integer, String> fzKl = new HashMap<Integer, String>();
        fzKl.put(1, "PkwG");
        fzKl.put(2, "PkwA");
        fzKl.put(3, "Lkw");
        fzKl.put(4, "LkwK");
        fzKl.put(5, "Bus");
        fzKl.put(6, "KfzNk");
        fzKl.put(7, "Pkw");
        fzKl.put(8, "LkwA");
        fzKl.put(9, "SattelKfz");
        fzKl.put(10, "Krad");
        fzKl.put(11, "Lfw");
        fzKl.put(32, "PkwÄ");
        fzKl.put(33, "LkwÄ");
        fzKl.put(64, "Kfz");

        // Fahrzeugklassen festlegen (gemäß TLS)
        String[] fzKlNamen = {"PkwG", "PkwA", "Lkw", "LkwK", "Bus", "KfzNk", "Pkw", "LkwA", "SattelKfz", "Krad", "Lfw", "PkwÄ", "LkwÄ", "Kfz"};

        int s, v, q;
        String text;

        // 13 Fahrzeugklassencodes einlesen
        Integer[] Fahrzeugklassen = new Integer[13];

        for (int i = 0; i < 13; i++) {
            Fahrzeugklassen[i] = deBlock.getDeData().readUnsignedByteAsInt();
        }

        // Geschwindigkeiten für die 13 Fahrzeugklassen einlesen
        for (int i = 0; i < 13; i++) {
            v = deBlock.getDeData().readUnsignedByteAsInt();
            if (fzKl.get(Fahrzeugklassen[i]) != null) {
                text = "v" + fzKl.get(Fahrzeugklassen[i]);
                data.getItem(text).getUnscaledValue("Wert").set(tls8BitToInt(v));
            }
        }

        // Standardabweichung für die 13 Fahrzeugklassen einlesen
        for (int i = 0; i < 13; i++) {
            s = deBlock.getDeData().readUnsignedByteAsInt();
            if (fzKl.get(Fahrzeugklassen[i]) != null) {
                text = "s" + fzKl.get(Fahrzeugklassen[i]);
                data.getItem(text).getUnscaledValue("Wert").set(tls8BitToInt(s));
            }
        }

        // v85 für Fahrzeugklasse 1 einlesen
        v = deBlock.getDeData().readUnsignedByteAsInt();
        text = "v85" + fzKl.get(1);
        data.getItem(text).getUnscaledValue("Wert").set(tls8BitToInt(v));

        // v85 für Fahrzeugklasse 17 () einlesen
        v = deBlock.getDeData().readUnsignedByteAsInt();
        text = "v85" + fzKl.get(7);
        data.getItem(text).getUnscaledValue("Wert").set(tls8BitToInt(v));

        // v85 für Fahrzeugklasse 32 () einlesen
        v = deBlock.getDeData().readUnsignedByteAsInt();
        text = "v85" + fzKl.get(32);
        data.getItem(text).getUnscaledValue("Wert").set(tls8BitToInt(v));

        // v85 für Fahrzeugklasse 64 einlesen
        v = deBlock.getDeData().readUnsignedByteAsInt();
        text = "v85" + fzKl.get(64);
        data.getItem(text).getUnscaledValue("Wert").set(tls8BitToInt(v));

        // qKfz für Fahrzeugklasse Kfz einlesen
        data.getItem("qKfz").getUnscaledValue("Wert").set(tls16BitToInt(deBlock.getDeData().readUnsignedLowHighByteAsInt()));

        // Fahrzeugmenge für die 13 Fahrzeugklassen einlesen
        for (int i = 0; i < 13; i++) {
            q = deBlock.getDeData().readUnsignedLowHighByteAsInt();
            if (fzKl.get(Fahrzeugklassen[i]) != null) {
                text = "q" + fzKl.get(Fahrzeugklassen[i]);
                data.getItem(text).getUnscaledValue("Wert").set(tls16BitToInt(q));
            }
        }

		/*
		  //  Geschwindigkeitsklassen einlesen
		  for (int j = 0; j < 13; j++) {
			  if (fzKl.get(Fahrzeugklassen[j]) != null) {
				  text = "q" + fzKl.get(Fahrzeugklassen[j]) + "GeschwKlasse";
				  count = deBlock.getDeData().readUnsignedByteAsInt();
				  array = data.getArray(text);
				  array.setLength(count);
  */

        //  Geschwindigkeitsklassen einlesen
        //  Geschwindigkeitsklassen qPkwÄ einlesen
        Integer[] geschwKl = new Integer[] {1, 2, 3, 4, 5, 7, 8, 9, 10, 11, 32, 33};
        for (Integer j : geschwKl) {
            text = "q" + fzKl.get(j) + "GeschwKlasse";
            count = deBlock.getDeData().readUnsignedByteAsInt();
            array = data.getArray(text);
            array.setLength(count);

            for (int i = 0; i < count; i++) {
                _qKlasse_ = deBlock.getDeData().readUnsignedLowHighByteAsInt();
                array.getItem(i).getUnscaledValue("Wert").set(tls16BitToInt(_qKlasse_));
                array.getItem(i).getItem("Status").getItem("Erfassung").getUnscaledValue("NichtErfasst").setText("Nein");
                array.getItem(i).getItem("Status").getItem("PlFormal").getUnscaledValue("WertMax").setText("Nein");
                array.getItem(i).getItem("Status").getItem("PlFormal").getUnscaledValue("WertMin").setText("Nein");
                array.getItem(i).getItem("Status").getItem("PlLogisch").getUnscaledValue("WertMaxLogisch").setText("Nein");
                array.getItem(i).getItem("Status").getItem("PlLogisch").getUnscaledValue("WertMinLogisch").setText("Nein");
                array.getItem(i).getItem("Status").getItem("MessWertErsetzung").getUnscaledValue("Implausibel").setText("Nein");
                array.getItem(i).getItem("Status").getItem("MessWertErsetzung").getUnscaledValue("Interpoliert").setText("Nein");
                array.getItem(i).getItem("Güte").getScaledValue("Index").set(1.0);
                array.getItem(i).getItem("Güte").getUnscaledValue("Verfahren").set(0);
            }
        }

        //Ergebnisdatensatz erzeugen...
        //resultData = new ResultData(_de.getDe(), dataDescription, _UnixZeit_, data, true);
        resultData = new ResultData(_de.getDe(), dataDescription, _UnixZeit_, data);
        return resultData;
    }

    /**
     * FG 1 : Typ: 250 Id: 4 Name: DaLVERDSVLage am MessQuerschnitt<br>
     * <p>
     * Beschreibung des Telegrammaufbaus und Inhalts siehe db_desc und snd-LEV-DAV von Heusch/Boesefeldt.
     *
     * @param deBlock         Der zu bearbeitende De-Block.
     * @param data            Der passende Datensatz für die Ziel Attributgruppe/Aspekt-Kombination.
     * @param dataDescription Die Datenbeschreibung für die Attributgruppe/Aspekt-Kombination.
     *
     * @return Ergebnisdatensatz mit den vom DeBlock in die entsprechende Attributgruppe konvertierten Daten.
     */
    public ResultData Fg001Typ250Id004AtgStörfallZustandAspStörfallVerfahrenRDSHB(DeBlock deBlock, Data data, DataDescription dataDescription) {
        ResultData resultData = null;

        // VLage-Byte aus Telegramm ----------------------------------------------------------------------------
        int _VLage_ = deBlock.getDeData().readUnsignedByteAsInt();

        data.getTimeValue("T").setSeconds(60);

        data.getUnscaledValue("Situation").set(_VLage_);

        data.getTimeValue("Horizont").setSeconds(0);// Analysewerte, kein Prognosehorizont
        data.getItem("Güte").getScaledValue("Index").set(1.0);// Keine Güteinfo von HB, auf 1.0 setzen
        data.getItem("Güte").getTextValue("Verfahren").setText("Standard");

        //Ergebnisdatensatz erzeugen...
        resultData = new ResultData(_de.getDe(), dataDescription, _eakConversionModul.getTimeStamp(), data);
        return resultData;
    }
}
